import os
import json
import logging
import functools


class Cache:
    args = None
    data = {}

    def __init__(self, args):
        self.args = args
        self.statedir = args.statedir or '.'
        self.filename = os.path.join(self.statedir, 'apicache.json')
        self.load()

    def load(self):
        if os.path.exists(self.filename):
            try:
                with open(self.filename, 'r') as filep:
                    self.data = json.loads(filep.read())
            except Exception as ex: #pylint: disable=broad-except
                logging.error(f'Error reading api cache {ex}')

    def save(self):
        try:
            with open(self.filename, 'w') as filep:
                filep.write(json.dumps(self.data))
        except Exception as ex: #pylint: disable=broad-except
            logging.error(f'Error writing api cache {ex}')

    def __contains__(self, item):
        return item in self.data

    def __setitem__(self, key, value):
        self.data[key] = value
        self.save()

    def __getitem__(self, key):
        return self.data.get(key)

    def purge(self):
        self.data = {}
        self.save()


def memoized(obj):
    '''decorate class methods with caching.
    Requires the class to have a dict-like attribute used as storage
    '''

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        if hasattr(args[0], 'cache') and args[0].cache:
            cache = args[0].cache
        else:
            return obj(*args, **kwargs)

        key = str(args[1:]) + str(kwargs)
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]
    return memoizer



class MgraphConnectorAuditlogsMixin:

    def list_directoryaudits(self, **kwargs):
        '''
        iterator over all audit log entries.
        '''
        return self.get_paged(f'/auditLogs/directoryAudits', **kwargs)



import sys
import logging
import json

import requests
import msal

from .tokencache import TokenCache
from ..utils import jdebug
from .deltaiterator import GenericDeltaIterator
from .exceptions import MgraphApiError
from .cache import Cache


class ApiArgs:
    cache = False
    purge_cache = False
    statedir = '.'
    debug = False
    v = False

    def __init__(self, args):
        if not args:
            return
        for key in ['cache', 'purge_cache', 'statedir', 'v', 'debug']:
            if key in args:
                setattr(self, key, getattr(args, key))


class MgraphBase:
    access_token = None
    params = None
    debug = False
    cache = None

    def __init__(self, params='config.json', database=None, args=None):
        self.params = json.load(open(params))
        self.database = database
        self.args = ApiArgs(args)

        keyfilename = self.params['private_key_file']
        if self.args.cache:
            self.cache = Cache(args)
            if self.args.purge_cache:
                self.cache.purge()

        # Create a preferably long-lived app instance which maintains a token cache.
        tokencache = TokenCache(self.args)
        self.app = msal.ConfidentialClientApplication(
            self.params["client_id"], authority=self.params["authority"],
            client_credential={"thumbprint": self.params["thumbprint"],
                               "private_key": open(keyfilename).read()
                               },
            token_cache=tokencache.cache
        )

    def enable_debug(self):
        self.args.debug = True

    def connect(self):
        # The pattern to acquire a token looks like this.
        result = None

        # Firstly, looks up a token from cache
        # Since we are looking for token for the current app, NOT for an end user,
        # notice we give account parameter as None.
        result = self.app.acquire_token_silent(
            self.params["scope"], account=None)

        if not result:
            logging.info(
                "No suitable token exists in cache. Let's get a new one from AAD.")
            result = self.app.acquire_token_for_client(
                scopes=self.params["scope"])

        if "access_token" in result:
            # Calling graph using the access token
            self.access_token = result['access_token']
            logging.info('Authorized.')
        else:
            logging.error(result.get("error"))
            logging.error(result.get("error_description"))
            # You may need this when reporting a bug
            logging.error(result.get("correlation_id"))
            raise MgraphApiError(result)

    def _assert_connected(self):
        if not self.access_token:
            self.connect()
        if not self.access_token:
            logging.error('No access_token available.')
            sys.exit(1)

    def _url(self, url):
        '''add endpoint to url if it has no complete protocol'''
        if url.startswith('/'):
            return self.params["endpoint"] + url
        return url

    def _generic_iterate(self, graph_data):
        '''private
        Support for paged results. See https://docs.microsoft.com/de-de/graph/paging
        '''
        while 'value' in graph_data:
            nextlink = None
            # "@odata.nextLink": "https://graph.microsoft.com/v1.0/users?$skiptoken=X%27...00%27"
            if "@odata.nextLink" in graph_data:
                nextlink = graph_data.get("@odata.nextLink")

            for val in graph_data['value']:
                yield val

            if nextlink:
                graph_data = self.get(nextlink)
            else:
                break

    def get(self, url, **kwargs):
        self._assert_connected()
        if kwargs:
            params = '&'.join([f"${k}={v}" for k, v in kwargs.items() if k])
            if params:
                params = '?' + params
        else:
            params = ''

        graph_data = requests.get(  # Use token to call downstream service
            self._url(url + params),
            headers={'Authorization': 'Bearer ' + self.access_token}, ).json()

        if self.args.debug:
            recordfile = 'log.json'
            with open(recordfile, 'a+') as fp:  # pylint: disable=invalid-name
                fp.write(f'# GET {url}\n')
                fp.write(json.dumps(graph_data, indent=2))
                fp.write('\n\n')

            logging.debug(f'GET {url} {params}')
            jdebug(graph_data, caller='get')

        if 'error' in graph_data:
            raise MgraphApiError(
                graph_data, f'MsGraph GET Error: {url}{params}')

        return graph_data

    def get_paged(self, url, **kwargs):
        '''
        iterator over all graph results. Supports @odata.nextLink paging
        '''
        return self.get_deltaiterator(url, **kwargs)

    def get_deltaiterator(self, url, deltalink=None, **kwargs):
        '''
        iterator over all graph results. Supports @odata.nextLink paging and maintains @odata.deltaLink for next call
        '''
        return GenericDeltaIterator(self, url, deltalink=deltalink, **kwargs)

    def get_binary(self, url, headers=None, fd=None):  # pylint: disable=invalid-name
        if headers is None:
            headers = {}
        headers.update({'Authorization': 'Bearer ' + self.access_token})
        response = requests.get(
            self._url(url),
            headers=headers)
        if self.args.debug:
            logging.debug(f'GET {url}')
        if response.status_code == requests.codes.ok:  # pylint: disable=no-member
            if fd:  # 4 MB, so we do not use too much memory here
                for chunk in response.iter_content(chunk_size=4096*1024):
                    logging.debug(f'chunk {len(chunk)}')
                    fd.write(chunk)
                return None
            return response.content
        else:
            logging.error(json.dumps(response.json()))
            raise MgraphApiError(response)
        return None

    def post(self, url, data=None):
        self._assert_connected()
        graph_data = requests.post(  # Use token to call downstream service
            self._url(url),
            json=data,
            headers={'Content-Type': 'application/json',
                     'Authorization': 'Bearer ' + self.access_token}, ).json()
        if self.args.debug:
            logging.debug(f'POST {url}')
            jdebug(graph_data, caller='post')
        if 'error' in graph_data:
            raise MgraphApiError(graph_data)
        return graph_data

    def post_paged(self, url, data=None):
        '''
        iterator over all graph results. Supports @odata.nextLink paging
        '''
        return self._generic_iterate(
            self.post(url, data=data)
        )

    def patch(self, url, data=None):
        self._assert_connected()

        graph_data = requests.patch(  # Use token to call downstream service
            self._url(url),
            json=data,
            headers={'Content-Type': 'application/json',
                     'Authorization': 'Bearer ' + self.access_token}, ).json()
        if self.args.debug:
            logging.debug(f'PATCH {url}')
            jdebug(graph_data, caller='patch')
        if 'error' in graph_data:
            raise MgraphApiError(graph_data)
        return graph_data

    def delete(self, url):
        self._assert_connected()

        response = requests.delete(  # Use token to call downstream service
            self._url(url),
            headers={'Content-Type': 'application/json',
                     'Authorization': 'Bearer ' + self.access_token}, )
        if self.args.debug:
            logging.debug(f'DELETE {url}')
            jdebug(response.json(), caller='delete')
        if response.status_code != 204:  # pylint: disable=no-member
            raise MgraphApiError(response.json())
        return response



# pylint: disable=logging-format-interpolation
import os
import atexit
import logging

import msal


class TokenCache:
    cache = None

    def __init__(self, args):
        cache_file = os.path.join(args.statedir, "tokencache.bin")
        cache = self.cache = msal.SerializableTokenCache() # pylint: disable=invalid-name
        if os.path.exists(cache_file):
            self.cache.deserialize(open(cache_file, "r").read())
            logging.debug(f'Loaded cache {cache_file}')
        atexit.register(lambda:
                        open(cache_file, "w").write(cache.serialize())
                        # Hint: The following optional line persists only when state changed
                        if cache.has_state_changed else None
                        )



from .cache import memoized


class MgraphConnectorGroupMixin:
    def list_groups(self, **kwargs):
        '''https://docs.microsoft.com/en-us/graph/api/resources/group?view=graph-rest-1.0'''
        url = f'/groups'
        search = kwargs.get('search')
        if search is not None:
            del kwargs['search']
            url += f"?$filter=startswith(displayName,'{search}')"
        return self.get_paged(url, **kwargs)

    @memoized
    def get_group(self, group_id):
        '''returns a group'''
        return self.get(f'/groups/{group_id}')

    def list_group_members(self, group_id):
        '''returns directoryObjects'''
        return self.get_paged(f'/groups/{group_id}/members')

    def get_directoryobject(self, object_id):
        return self.get(f'/directoryObjects/{object_id}')

    def list_group_owners(self, group_id):
        return self.get_paged(f'/groups/{group_id}/owners')

    def get_group_drive(self, group_id):
        return self.get(f'/groups/{group_id}/drive')

    def get_group_drives(self, group_id):
        return self.get(f'/groups/{group_id}/drives')



import logging
from urllib.parse import urlparse, parse_qs


class MgraphConnectorDriveMixin:
    root_item = None
    drive_id = None

    def open(self, drive_id):
        self.drive_id = drive_id

    def download(self, item_id, fd): #pylint: disable=invalid-name
        return self.get_binary(f'/drives/{self.drive_id}/items/{item_id}/content', fd=fd)

    def get_root(self):
        root = self.get(f'/drives/{self.drive_id}/root')
        self.root_item = root
        return root

    def get_driveitem(self, item_id):
        return self.get(f'/drives/{self.drive_id}/items/{item_id}')

    def __parse_deltaurl(self, url):
        url_parts = urlparse(url)
        query = parse_qs(url_parts.query)
        token = None
        if 'token' in query:
            token = query['token']
        if token:
            token = token[0]
        return token

    def __get_delta_chunk(self, drive_id, token=None):
        url = f'/drives/{drive_id}/root/delta'
        if token is not None:
            url += f"?token='{token}'"
        logging.debug(f'get_delta_chunk({url})')
        return self.get(url)

    def drive_delta_iterator(self, ignore_state=False):
        '''iterator over changes'''
        tokens_seen = {}
        state_db = self.database

        state = state_db.sync.find_one(drive_id=self.drive_id)
        logging.debug(f'State: {state}')
        if state and not ignore_state:
            token = state['token']
            logging.info(f'Resuming from {token}')
        else:
            token = None
            logging.info('New sync')

        while True:
            try:
                delta = self.__get_delta_chunk(self.drive_id, token=token) # FIXME fails on deleted drives
            except Exception as ex:
                logging.error(f"get delta exception {ex}")
                return
            # status 200 ok
            # status 410 gone:
            #   resyncChangesApplyDifferences Replace any local items with the server's version (including deletes) if you're sure that the service was up to date with your local changes when you last sync'd. Upload any local changes that the server doesn't know about.
            #   resyncChangesUploadDifferences Upload any local items that the service did not return, and upload any files that differ from the server's version (keeping both copies if you're not sure which one is more up-to-date).
            #   check Location header containing a new nextLink

            deltanext = delta.get('@odata.nextLink')
            deltalink = delta.get('@odata.deltaLink')
            value = delta.get("value", [])
            for v in value: # pylint: disable=invalid-name
                yield v

            if deltalink: # last page in a set
                token = self.__parse_deltaurl(deltalink)
                logging.debug(f'Got deltalink {token} {deltalink}')
                state_db.sync.upsert(
                    dict(drive_id=self.drive_id, deltalink=deltalink, token=token),
                    ['drive_id']
                )
                break

            if deltanext: # there are more pages
                logging.debug(f'Got deltanext {deltanext}')
                # https://graph.microsoft.com/v1.0
                # /drives/b!edsxy_Z5eEGwX_1XwPTeTMw3qRNusSdKoxH1k8HgnQ_nl7H6J8PfRLZq170BWA32
                #   /root/delta?token=...
                token = self.__parse_deltaurl(deltanext) # FIXME do not parse, use the url as it is!
                if token in tokens_seen:
                    logging.error(f'Token already seen {token}')
                    break
                tokens_seen[token] = True
            else:
                token = None
            if not token:
                break



class MgraphConnectorDeletedMixin:
    def list_deleted_users(self, **kwargs):
        '''
        iterator over all deleted graph users.
        '''
        return self.get_paged(f'/directory/deletedItems/microsoft.graph.user', **kwargs)

    def list_deleted_groups(self, **kwargs):
        '''
        iterator over all deleted graph groups.
        '''
        return self.get_paged(f'/directory/deletedItems/microsoft.graph.group', **kwargs)

    def list_deleted_applications(self, **kwargs):
        '''
        iterator over all deleted graph applications.
        '''
        return self.get_paged(f'/directory/deletedItems/microsoft.graph.application', **kwargs)

    def get_deleted_item(self, object_id):
        '''
        get a deleted item by object_id
        '''
        return self.get(f'/directory/deletedItems/{object_id}')




class MgraphApiError(Exception):
    code = None
    graph_message = None
    inner_error = None

    def __init__(self, graph_data, message=None):
        super(MgraphApiError, self).__init__(graph_data)
        self.graph_data = graph_data
        self.message = message
        # see https://docs.microsoft.com/en-us/graph/errors
        error = self.graph_data.get('error')
        if error:
            self.code = error.get('code', 'unknown')
            self.graph_message = error.get('message', '?')
            self.inner_error = error.get('innerError')

    def __str__(self):
        if self.code:
            return f'''{self.code}: {self.graph_message}
    Inner error: {self.inner_error}
    Message: {self.message}'''
        return super(MgraphApiError, self).__str__()



class MgraphConnectorUserContactMixin:
    def mk_url(self, uid=None, folder=None):
        url = '/contacts'
        if uid is not None:
            url = f'/users/{uid}/contacts'
        if folder is not None:
            url = f'/users/{uid}/contactFolders/{folder}/contacts'
        return url

    def get_contact(self, uid=None, folder=None, contact_id=None):
        url = self.mk_url(uid, folder)
        url += f'/{contact_id}'
        return self.get(url)

    def list_contacts(self, uid=None, folder=None):
        url = self.mk_url(uid, folder)
        return self.get(url)

    def create_contact(self, uid=None, folder=None, data=None):
        url = self.mk_url(uid, folder)
        return self.post(url, data)

    def update_contact(self, uid=None, cid=None, data=None):
        url = f'/users/{uid}/contacts/{cid}'
        return self.patch(url, data)

    def delete_contact(self, uid=None, cid=None):
        url = f'/users/{uid}/contacts/{cid}'
        return self.delete(url)



import logging
import json
import os


def jdump(data, caller=None):
    try:
        logging.info(f'{caller}\n{json.dumps(data, indent=2)}')
    except:
        logging.info(f'{caller}\n{repr(data)}')

def jdebug(data, caller=None):
    try:
        logging.debug(f'{caller}\n{json.dumps(data, indent=2)}')
    except:
        logging.debug(f'{caller}\n{repr(data)}')



import os

from .api import MgraphApi
from .api.deltaiterator import GenericDeltaIterator

HOME = os.environ.get('HOME')
CFGDIR = os.path.join(HOME, '.config/wmssync')
STATEDIR = os.path.join(HOME, '.local/share/wmssync')

def main():
    print(f'testing {MgraphApi} {GenericDeltaIterator}')
    paramsfile = os.path.join(CFGDIR, 'config.json')
    api = MgraphApi(params=paramsfile)
    # url = 'https://graph.microsoft.com/v1.0/groups/delta?$select=displayName,description,members'
    url = 'https://graph.microsoft.com/v1.0/users/delta?$select=displayName,userPrincipalName'
    iterator = api.get_deltaiterator(url, deltalink=None)
    for item in iterator:
        print(item)
    print(f"next iteration: {iterator.get_next_url()}")
    # for user in api.list_users():
    #     print(user)

if __name__=='__main__':
    main()


from .cache import memoized


class MgraphConnectorUserMixin:

    def list_users(self, **kwargs):
        '''
        iterator over all graph users.
        '''
        return self.get_paged(f'/users', **kwargs)

    @memoized
    def get_user(self, user_id):
        '''
        iterator over all graph users.
        '''
        # return [self.get(f'/users/f05885f7-1690-4ac3-99c1-1f69d874bc3c')]
        return self.get(f'/users/{user_id}')

    def get_user_licensedetails(self, user_id):
        return self.get(f'/users/{user_id}/licenseDetails')['value']

    def get_user_member_groups(self, user_id):
        return self.post_paged(f'/users/{user_id}/getMemberGroups',
                               data={'securityEnabledOnly': False})

    def delete_user(self, user_id):
        return self.delete(f'/users/{user_id}')



#!/usr/bin/env python
# upload to pypi: python3 -m twine upload --repository testpypi dist

import codecs
import os.path

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setup(
    name="wmgraph",
    version=get_version("wmgraph/__init__.py"),
    author="Patrick Atamaniuk, wibas GmbH",
    author_email="patrick.atamaniuk@wibas.com",
    description="Microsoft Graph base library",
    long_description="""Microsoft Graph base library

    api and utility methods to use MS Graph functionality
    """,
    url="https://www.wibas.com/",
    license="MIT",
    packages=find_packages(),
    # scripts = ["wmssync"],
    entry_points={
        'console_scripts': ['wmssync=wmssync.cli:main']
    },
    install_requires=[
        "requests>=2,<3",
        "msal>=0,<2",
        "quickxorhash",
    ],
)



import json
import logging

from ..utils import jdump

from .base import MgraphBase
from .drive import MgraphConnectorDriveMixin
from .group import MgraphConnectorGroupMixin
from .usercontact import MgraphConnectorUserContactMixin
from .site import MgraphConnectoSiteMixin
from .user import MgraphConnectorUserMixin
from .deleted import MgraphConnectorDeletedMixin
from .auditlogs import MgraphConnectorAuditlogsMixin

class MgraphApi(
        MgraphBase,
        MgraphConnectorDriveMixin,
        MgraphConnectorGroupMixin,
        MgraphConnectorUserMixin,
        MgraphConnectoSiteMixin,
        MgraphConnectorUserContactMixin,
        MgraphConnectorAuditlogsMixin,
        MgraphConnectorDeletedMixin,
        ):
    pass

class MgraphApiMockX(MgraphApi):
    def __init__(self, params='config.json', database=None, args=None): # pylint: disable=super-init-not-called
        self.database = database
        self.args = args

class MgraphApiMock(MgraphApi):
    def get(self, url, params=None, cache_id=None):
        '''cache and mock'''
        item_id = f"{url}?{params}"
        res = self.database.requests.find_one(item_id=item_id)
        if res:
            res = json.loads(res['data'])
            if self.args.v:
                jdump(res, caller=__file__)
            logging.debug(f'cached response {item_id}')
        else:
            res = super(MgraphApiMock, self).get(url, params=params)
            self.database.requests.insert(dict(item_id=item_id, data=json.dumps(res)))
            logging.debug(f'cached request {item_id}')
        return res



class MgraphConnectoSiteMixin:
    def get_root_site(self):
        return self.get('/sites/root')

    def get_site(self, site_id):
        return self.get(f'/sites/{site_id}')

    def get_site_drive(self, site_id):
        return self.get(f'/sites/{site_id}/drive')

    def get_site_drives(self, site_id):
        res = self.get(f'/sites/{site_id}/drives')
        if 'value' in res:
            return res['value']
        return []

    def search_sites(self, site_name):
        res = self.get(f'/sites?search={site_name}')
        if 'value' in res:
            return res['value']
        return []

    def get_subsites(self, site_id):
        return self.get(f'/sites/{site_id}/sites')

    def walk_sites(self, site=None):
        if site is None:
            site = self.get_root_site()
            self.sites = {}
            self.sites[site['id']] = site
        subsites = self.get_subsites(site['id'])
        # jdump(subsites, caller=__file__)
        if 'value' in subsites:
            for subsite in subsites['value']:
                self.sites[subsite['id']] = subsite
                self.walk_sites(site=subsite)
        return self.sites




class GenericDeltaIterator:
    def __init__(self, api, url=None, deltalink=None, **kwargs):
        '''
        Delta iterator.
        If deltalink is given, url is not used and the iterator resumes from that state.

        Example:
            url = 'https://graph.microsoft.com/v1.0/users/delta?$select=displayName,userPrincipalName'
            iterator = GenericDeltaIterator(api, url)
            for item in iterator:
                print(item)
            print(f"next iteration: {iterator.get_next_url()}")
        '''
        self.api = api
        self.url = deltalink or url
        self.kwargs = kwargs
        self.deltalink = deltalink

    def __iter__(self):
        return self._generic_iterate()

    def get_next_url(self):
        return self.deltalink

    def _generic_iterate(self):
        '''private
        Support for paged results. See https://docs.microsoft.com/de-de/graph/paging
        Saves deltalink for delta operations
        '''
        graph_data = self.api.get(self.url, **self.kwargs)
        while 'value' in graph_data:
            deltanext = graph_data.get('@odata.nextLink')
            self.deltalink = graph_data.get('@odata.deltaLink')

            for val in graph_data['value']:
                yield val

            if deltanext:
                graph_data = self.api.get(deltanext)
            else:
                break
